{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "eb866408-5568-48d6-a3c4-b4f788397aee",
   "metadata": {},
   "source": [
    "# Modern arrays\n",
    "\n",
    "Arrays are useful to store large sets of data that share the same data type.\n",
    "Arrays are very fast to access and modify data, and are useful in several cases,\n",
    "like\n",
    "\n",
    "-   Reading and storing data.\n",
    "-   Applying the same operation to a large collection of data.\n",
    "-   Store data in a systematic way instead of having many variables.\n",
    "-   Model vectors, matrices, tensors.\n",
    "-   Model vectorial and matrix operations.\n",
    "-   Solve large linear problems, linear algebra problems, compute eigen values and\n",
    "    eigen vectors.\n",
    "-   Solve ODE systems.\n",
    "-   Apply finite differences and finite elements.\n",
    "-   Vectorize operations and use HPC.\n",
    "-   Model data structs, from physical systems to graphical structs and so on.\n",
    "-   Manipulate memory contents and program close to the metal.\n",
    "-   &#x2026;\n",
    "\n",
    "There are many more data structs or [containers](https://en.cppreference.com/w/cpp/container): lists, dictionaries(maps), queue, binary trees, etc (see\n",
    "<https://www.cs.usfca.edu/~galles/visualization/Algorithms.html>). Arrays, in\n",
    "particular, are very fast to access and retrieve data from memory. But they are\n",
    "not fast if you need to insert and delete data in the middle. In general, for\n",
    "scientific computing, arrays are the default data struct, but you need always to\n",
    "consider which one would be the best for your case.\n",
    "\n",
    "Since c++11 it is possible to use modern data structs that are equivalent to the\n",
    "primitive data arrays but give a better user interface, also handling\n",
    "automatically the memory. Although there are two analog structures, [std::array](https://en.cppreference.com/w/cpp/container/array)\n",
    "and [std::vector](https://en.cppreference.com/w/cpp/container/vector) (and in the future we will use [std::valarray](https://en.cppreference.com/w/cpp/numeric/valarray)), we will focus\n",
    "mainly on the `std::vector` since it handles dynamic memory (`std::array` is for\n",
    "static arrays). The key here is that these data structures create a continuous\n",
    "block memory block that can be accessed very fast: https://hackingcpp.com/cpp/std/sequence_containers.html\n",
    "\n",
    "<div style=\"text-align: center;\">\n",
    "    <img src=\"https://hackingcpp.com/cpp/std/vector.png\" alt=\"Image Description\" width=\"900\">\n",
    "    <figcaption>From: \"https://hackingcpp.com/cpp/std/vector.png\"</figcaption>\n",
    "</div>"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "01bd1d48-06f3-48f3-a7ff-72c551942d13",
   "metadata": {},
   "source": [
    "The following example shows the basic features for both of them and also a new\n",
    "kind of `for`. It is recommended to run this on a graphical debugger , either\n",
    "locally in emacs or visual code, or online on the python tutor or [onlinegdb](https://www.onlinegdb.com/), at\n",
    "the link <https://onlinegdb.com/CKeYkLEiz> .\n",
    "\n",
    "```cpp\n",
    "#include <iostream>\n",
    "#include <array>\n",
    "#include <vector>\n",
    "\n",
    "int main(void)\n",
    "{\n",
    "  std::array<double, 10> data1;\n",
    "  data1[5] = 9.87;\n",
    "  data1[0] = -7.8765;\n",
    "  data1[9] = 1000.343;\n",
    "  std::cout << data1[5] << \"\\n\";\n",
    "  for (auto val : data1) {\n",
    "    std::cout << val << \"\\n\";\n",
    "  }\n",
    "\n",
    "  std::vector<double> data2;\n",
    "  data2.resize(4); // size = 4, capacity 4\n",
    "  data2[3] = 9.87;\n",
    "  data2[0] = -7.8765;\n",
    "  data2[1] = 1000.343;\n",
    "  std::cout << data2[2] << std::endl;\n",
    "  for (auto & val : data2) { // reference\n",
    "    val *= 2; // modify the value\n",
    "    std::cout << val << \"\\n\";\n",
    "  }\n",
    "  data2.resize(5); // size = 5, capacity = 8 (double the initial, to avoid frequent mallocs)\n",
    "  data2.push_back(0.987); // size = 6, capacity = 8\n",
    "  data2.push_back(12343.987); // size = 7, capacity = 8\n",
    "  data2.push_back(-0.000987); // size = 8, capacity = 8\n",
    "  data2.push_back(3.986544); // size = 9, capacity = 16\n",
    "\n",
    " return 0;\n",
    "}\n",
    "```\n",
    "\n",
    "As you can see, a vector is a dynamc array of homogeneous contigous elements.\n",
    "\n",
    "\n",
    "<a id=\"orga3d147a\"></a>"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5a8f4b19-c791-4e36-bdda-2dcb60cceae6",
   "metadata": {},
   "source": [
    "## Initial exercises\n",
    "\n",
    "1.  Find the maximum size you can use for a `std::array` and a `std::vector` .\n",
    "    Are they the same?\n",
    "2.  Print the address of the first and second elements in a `std::array` and\n",
    "    `std::vector`. Are they contigous? To print the memory address of the first\n",
    "    element use the `&` operator as `&data[0]`\n",
    "3.  Use the algorithm [for<sub>each</sub>](https://en.cppreference.com/w/cpp/algorithm/for_each) with a lambda function to divide by two each\n",
    "    element in an array. Use also for each to print the elements and their memory\n",
    "    address."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "62554028-925c-47bd-8984-646f359169cd",
   "metadata": {},
   "source": [
    "<a id=\"orgb9267a4\"></a>\n",
    "\n",
    "## Using `std::array` : Fixed size, goes to the stack, automatic for\n",
    "\n",
    "```cpp\n",
    "// g++ -g -fsanitize=address\n",
    "#include <iostream>\n",
    "#include <array>\n",
    "\n",
    "int main(void)\n",
    "{\n",
    "  const int N = 10;\n",
    "  //double data[N] {0};\n",
    "  std::array<double, N> data;\n",
    "\n",
    "  std::cout << \"size: \" << data.size() << std::endl;\n",
    "  std::cout << \"max_size: \" << data.max_size() << std::endl;\n",
    "  std::cout << \"&data[0]: \" << &data[0] << std::endl;\n",
    "  std::cout << \"&data[1]: \" << &data[1] << std::endl;\n",
    "\n",
    "  //std::cout << data[-1] << std::endl; // detected by sanitizers address\n",
    "\n",
    "// initialize the array with even numbers\n",
    "  for(int ii = 0; ii < N; ++ii) {\n",
    "    data[ii] = 2*ii;\n",
    "  }\n",
    "\n",
    "\n",
    "// print the array\n",
    "  for(const auto & val : data) {\n",
    "    std::cout << val << \"  \";\n",
    "  }\n",
    "  std::cout << \"\\n\";    \n",
    "\n",
    "// compute the mean\n",
    "  double sum = 0.0;\n",
    "  for(auto val : data) {\n",
    "    sum += val;\n",
    "  }\n",
    "  std::cout << \"Average = \" << sum/data.size() << std::endl;\n",
    "\n",
    "  return 0;\n",
    "}\n",
    "```\n",
    "You might visualize this in the python/c++ tutor.\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a39d1673-ebac-472c-accf-2fc75b4c89cb",
   "metadata": {},
   "source": [
    "And we can simplified the code even more if we use the standard library:\n",
    "```c++\n",
    "// g++ -g -fsanitize=address\n",
    "#include <numeric>\n",
    "#include <algorithm>\n",
    "#include <iostream>\n",
    "#include <array>\n",
    "\n",
    "int main(void)\n",
    "{\n",
    "  const int N = 10;\n",
    "  //double data[N] {0};\n",
    "  std::array<double, N> data;\n",
    "\n",
    "  std::cout << \"size: \" << data.size() << std::endl;\n",
    "  std::cout << \"max_size: \" << data.max_size() << std::endl;\n",
    "  std::cout << \"&data[0]: \" << &data[0] << std::endl;\n",
    "  std::cout << \"&data[1]: \" << &data[1] << std::endl;\n",
    "\n",
    "  //std::cout << data[-1] << std::endl; // detected by sanitizers address\n",
    "\n",
    "// initialize the array with even numbers\n",
    "  int ii = 0 ;\n",
    "  auto init  = [&ii](double & x){ x = 2*ii; ii++; };\n",
    "  std::for_each(data.begin(), data.end(), init);\n",
    "\n",
    "// print the array\n",
    "  auto print = [](const int & x) { std::cout << x << \"  \" ; };\n",
    "  std::for_each(data.begin(), data.end(), print);\n",
    "  std::cout << \"\\n\";\n",
    "\n",
    "// compute mean\n",
    "  double avg = std::accumulate(data.begin(), data.end(), 0.0)/data.size();\n",
    "  std::cout << \"Average = \" << avg << std::endl;\n",
    "}\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "219d2273-32fd-4633-8120-bd455eb72f53",
   "metadata": {},
   "source": [
    "<a id=\"org942e5a1\"></a>\n",
    "\n",
    "### Exercises\n",
    "\n",
    "1.  Find the largest `std::array` size you can get in your system. Is it related\n",
    "    to the ram size? check the command\n",
    "    \n",
    "        ulimit -s\n",
    "2.  Write a program that computes the dot product between two `std::arrays`. You\n",
    "    can use either indexes or the `inner_product` or `transform_reduce`\n",
    "    algorithms.\n",
    "3.  Can you read the `std::array` size from the command line using `argv`? try\n",
    "    it.\n",
    "\n",
    "\n",
    "<a id=\"org02d3539\"></a>"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "73ce60ce-6346-4268-b1ac-03393811f134",
   "metadata": {},
   "source": [
    "## Using `std::vector` : Dynamic size, uses the heap\n",
    "\n",
    "Vectors allow to use dynamic size arrays very easily and are the recommend data\n",
    "structure for our current math vector and matrix applications. You can see more\n",
    "info here:\n",
    "\n",
    "-   <https://hackingcpp.com/cpp/std/sequence_containers.html#vector>\n",
    "-   <https://hackingcpp.com/cpp/std/vector_intro.html>\n",
    "-   <https://hackingcpp.com/cpp/std/vector.html>\n",
    "-   <https://www.learncpp.com/cpp-tutorial/an-introduction-to-stdvector/>\n",
    "-   <https://www.learncpp.com/cpp-tutorial/stdvector-capacity-and-stack-behavior/>\n",
    "\n",
    "For now, let's just declare a vector\n",
    "\n",
    "```c++\n",
    "// g++ -g -fsanitize=address\n",
    "#include <algorithm>\n",
    "#include <numeric>\n",
    "#include <iostream>\n",
    "#include <vector>\n",
    "\n",
    "int main(int argc, char **argv)\n",
    "{\n",
    "  const int N = std::atoi(argv[1]);\n",
    "  std::vector<double> data;\n",
    "  data.reserve(5); //affects capacity, can store 15 numbers but its size is still 0\n",
    "  data.resize(N); // real size, capacity >= N\n",
    "\n",
    "  std::cout << \"size: \" << data.size() << std::endl;\n",
    "  std::cout << \"capacity: \" << data.capacity() << std::endl;\n",
    "  std::cout << \"max_size: \" << data.max_size() << std::endl;\n",
    "  std::cout << \"&data[0]: \" << &data[0] << std::endl;\n",
    "  std::cout << \"&data[1]: \" << &data[1] << std::endl;\n",
    "\n",
    "  //std::cout << data[-1] << std::endl; // detected by sanitizers address\n",
    "\n",
    "// initialize the array with even numbers\n",
    "  int ii = 0;\n",
    "  auto init  = [&ii](double & x){ x = 2*ii; ii++; };\n",
    "  std::for_each(data.begin(), data.end(), init);\n",
    "\n",
    "// print the array\n",
    "  auto print = [](const double & x){ std::cout << x << \"  \";};\n",
    "  std::for_each(data.begin(), data.end(), print);\n",
    "  std::cout << \"\\n\";\n",
    "\n",
    "// compute mean\n",
    "  std::cout << \"Average: \" << std::accumulate(data.begin(), data.end(), 0.0)/data.size() << \"\\n\";\n",
    "\n",
    "// Add more data!\n",
    "  std::cout << \"size: \" << data.size() << \", capacity: \" << data.capacity() << std::endl;\n",
    "  for (auto x : {200.0987, 300.987, 400.09754}){\n",
    "    data.push_back(x);\n",
    "    std::cout << \"size: \" << data.size() << \", capacity: \" << data.capacity() << std::endl;\n",
    "  }\n",
    "  std::for_each(data.begin(), data.end(), print);\n",
    "  std::cout << \"\\n\";\n",
    "\n",
    "  return 0;\n",
    "}\n",
    "```\n",
    "\n",
    "<a id=\"org581880d\"></a>"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "92d370bb-b79a-476c-b5b1-c98fdd0070c4",
   "metadata": {},
   "source": [
    "## Exercises\n",
    "\n",
    "Always try to use the STL, and also use functions to implement the following:\n",
    "\n",
    "1.  Compute the norm of a vector.\n",
    "2.  Implement the Gauss 7 point method, storing the weights and the coordinates points in a vector.\n",
    "3.  Compute the dot product between two vectors. Use `std::inner_product`.\n",
    "4.  Assume that a vector represents the coefficients of a n-degree polynomial.\n",
    "    Compute the derivative of the polynomial and save the coefficients in another\n",
    "    vector.\n",
    "    ```cpp\n",
    "        data = {1 , 3, 4.5}; // 1*x^0 + 3*x^1 + 4.5*x^2\n",
    "        newdata = {3, 9.0}; // 0 + 3*x^0 + 2*4.5*x^1\n",
    "    ```\n",
    "5.  Compute the min of a vector.\n",
    "6.  Compute the max of a vector.\n",
    "7.  Compute the argmin (index of the min) of a vector\n",
    "8.  Compute the argmax (index of the max) of a vector\n",
    "9.  Compute the pnorm\n",
    "    \n",
    "    \\begin{equation}\n",
    "    \\left(\\sum_{i} x_{i}^{p}\\right)^{1/p} \\ .\n",
    "    \\end{equation}\n",
    "10. Read the contents of a large file into a vector. Compute and print the mean,\n",
    "    the median, the 25, 50, 75 percentiles, the min value, and the max value.\n",
    "11. Use the standard algorithm sort to sort a data array.\n",
    "12. Fill a vector with random numbers\n",
    "    ```cpp\n",
    "        // llenar un vector con numeros aleatorios\n",
    "        \n",
    "        #include <iostream>\n",
    "        #include <string>\n",
    "        #include <vector>\n",
    "        #include \"random_vector.h\"\n",
    "        \n",
    "        int main(int argc, char **argv) {\n",
    "        \n",
    "          // read variables\n",
    "          if (argc != 2) {\n",
    "            std::cerr << \"Usage: \" << argv[0] << \" N\" << std::endl;\n",
    "            std::cerr << \"N: size of the vector\" << std::endl;\n",
    "            return 1;\n",
    "          }\n",
    "          const int N = std::stoi(argv[1]);\n",
    "        \n",
    "          // set the vector\n",
    "          std::vector<double> data;\n",
    "          data.resize(N);\n",
    "        \n",
    "          // fill the vector\n",
    "          fill_randomly(data);\n",
    "        \n",
    "          // print the vector\n",
    "          std::string fname = \"data.txt\";\n",
    "          print(data, fname);\n",
    "        \n",
    "          return 0;\n",
    "        }\n",
    "    ```\n",
    "13. Create a vector of ints, and then use `std::shuffle` to randomly shuffle it.\n",
    "14. Use a vector to compute the counter histogram for some data read from a\n",
    "    file.\n",
    "15. Histogram from random numbers, saving to file. Write a program that\n",
    "    generates N random numbers , with the weibull distribution and a given seed.\n",
    "    Both N, the seed, and the weibull parameters must read from the command\n",
    "    line. Also, you must compute the histogram and print it to a file."
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
